// =====================================================================
//
// xmlrpc_client.cxx
//
// connect to flarq xmlrpc server
//
// Copyright (C) 2007-2009
//		Dave Freese, W1HKJ
//
// This file is part of fldigi.
//
// Fldigi is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Fldigi is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with fldigi.  If not, see <http://www.gnu.org/licenses/>.
// =====================================================================

#include <config.h>

#include <iostream>
#include <cmath>
#include <cstring>
#include <vector>
#include <list>
#include <stdlib.h>

#include <FL/Fl.H>
#include <FL/filename.H>
#include <FL/fl_ask.H>

#include "flmsg.h"
#include "threads.h"
#include "debug.h"

#include "xmlrpc_client.h"
#include "xmlrpcpp/XmlRpc.h"

using namespace XmlRpc;
using namespace std;

static int xmlrpc_arq_verbosity = 0;

//======================================================================
// flarq xmlrpc client
//======================================================================

pthread_t *flarq_thread = 0;
pthread_mutex_t mutex_flarq = PTHREAD_MUTEX_INITIALIZER;

bool flarq_get_xcvr();
void flarq_connection();

void connect_to_flarq();

XmlRpcClient *flarq_client = (XmlRpcClient *)0;

//----------------------------------------------------------------------
// declared as extern
//----------------------------------------------------------------------
bool connected_to_flarq = false;
bool send_using_flarq = false;
//----------------------------------------------------------------------

static double timeout = 2.0;
//======================================================================
#define FLARQ_GET_VERSION "flarq.get_version"
#define FLARQ_GET_STATE   "flarq.get_state"
#define FLARQ_RCVD_TEXT   "flarq.rcvd_text"
#define FLARQ_SEND_TEXT   "flarq.send_text"

std::string flarq_get_version()
{
	guard_lock flarq_lock(&mutex_flarq);
	XmlRpcValue result;
	if (flarq_client->execute(FLARQ_GET_VERSION, XmlRpcValue(), result, timeout) ) {
		std::string arq_version = std::string(result);
printf("ARQ version %s\n", arq_version.c_str());
		return arq_version;
	} else {
		connected_to_flarq = false;
		LOG_ERROR("%s failed", FLARQ_GET_VERSION);
		return "";
	}
}

int flarq_get_state()
{
	guard_lock flarq_lock(&mutex_flarq);
	XmlRpcValue result;
	if (flarq_client->execute(FLARQ_GET_STATE, XmlRpcValue(), result, timeout) ) {
		int arq_state = int(result);
		LOG_INFO("ARQ state %02X", arq_state);
		return arq_state;
	} else {
		connected_to_flarq = false;
		LOG_ERROR("%s failed", FLARQ_GET_STATE);
		return XMLARQ_UNCONNECTED;
	}
}

std::string flarq_rcvd_text()
{
	guard_lock flarq_lock(&mutex_flarq);
	XmlRpcValue result;
	if (flarq_client->execute(FLARQ_RCVD_TEXT, XmlRpcValue(), result, timeout) ) {
		std::string arq_rcvd_text = std::string(result);
		return arq_rcvd_text;
	} else {
		connected_to_flarq = false;
		LOG_ERROR("%s failed", FLARQ_RCVD_TEXT);
		return "";
	}
}

void flarq_send_text(std::string txt)
{
	try {
		XmlRpcValue res;
		XmlRpcValue sending = txt;
		flarq_client->execute(FLARQ_SEND_TEXT, sending, res);
	} catch (const XmlRpc::XmlRpcException& e) {
		LOG_ERROR("%s", e.getMessage().c_str());
	}
}

//======================================================================
// xmlrpc read polling thread
//======================================================================
static bool run_flarq_thread = true;
static int poll_interval = 100; // 1 second

void flarq_connection()
{
	XmlRpcValue noArgs, result;
	try {
		bool ret;
		{
			guard_lock flarq_lock(&mutex_flarq);
			ret = flarq_client->execute("system.listMethods", noArgs, result, timeout);
		}
		if (ret) {
			int nargs = result.size();
			string method_str = "\nMethods:\n";
			for (int i = 0; i < nargs; i++)
				method_str.append("    ").append(result[i]).append("\n");
			LOG_INFO("%s", method_str.c_str());
			connected_to_flarq = true;
			poll_interval = 100; // every second
			flarq_get_version();
		} else {
			LOG_INFO("%s", "Waiting for flarq");
			connected_to_flarq = false;
			poll_interval = 500; // every 5 seconds
		}
	} catch (...) {
		LOG_ERROR("%s", "failure in flarq_client");
	}
}

void connect_to_flarq()
{
	XmlRpc::setVerbosity(xmlrpc_arq_verbosity);
	if (flarq_client) {
		delete flarq_client;
		flarq_client = (XmlRpcClient *)0;
	}
	try {
		flarq_client = new XmlRpcClient(
				"127.0.0.1", 7422);
		LOG_INFO("created flarq xmlrpc client 127.0.0.1:7422");
		flarq_connection();
	} catch (...) {
		LOG_ERROR("Cannot create flarq xmlrpc client 127.0.0.1:7422");
	}
}

void * flarq_thread_loop(void *d)
{
	int state = XMLARQ_UNCONNECTED;
	while(run_flarq_thread) {
		for (int i = 0; i < poll_interval; i++) {
			if (!run_flarq_thread) {
				LOG_INFO("Exiting thread - 1");
				return NULL;
			}
			MilliSleep(10);
		}

		if (!run_flarq_thread) break;

		if (!flarq_client)
			connect_to_flarq();
		if (!connected_to_flarq) flarq_connection();

		state = flarq_get_state();
		if (state == XMLARQ_RCVD) {
			std::string rxtext = flarq_rcvd_text();
			extract_text(rxtext, NULL);
		}
	}
	LOG_INFO("Exiting thread - 2");
	return NULL;
}

void start_flarq_thread()
{
	flarq_thread = new pthread_t;
	poll_interval = 500;
	if (pthread_create(flarq_thread, NULL, flarq_thread_loop, NULL)) {
		LOG_ERROR("%s", "flarq_thread create");
		exit(EXIT_FAILURE);
	}
}

void stop_flarq_thread()
{
	LOG_INFO("%s", "stopping flarq thread");
	flarq_client->close();
	pthread_mutex_lock(&mutex_flarq);
		run_flarq_thread = false;
	pthread_mutex_unlock(&mutex_flarq);
	pthread_join(*flarq_thread, NULL);
	LOG_INFO("%s", "flarq thread closed");
}

void reconnect_to_flarq()
{
	flarq_client->close();
	pthread_mutex_lock(&mutex_flarq);
	delete flarq_client;
	flarq_client = (XmlRpcClient *)0;
	connected_to_flarq = false;
	pthread_mutex_unlock(&mutex_flarq);
}

